import copy
import logging
import math
from typing import Optional, Dict, Union, Any, List, Callable

import numpy as np
import pandas as pd
from sklearn.preprocessing import QuantileTransformer

from econml.cate_interpreter import SingleTreeCateInterpreter  # noqa F401
from dowhy.causal_estimator import CausalEstimate
from dowhy import CausalModel


from causaltune.thirdparty.causalml import metrics
from causaltune.erupt import ERUPT
from causaltune.utils import treatment_values, psw_joint_weights

import dcor

# Imports for CODEC
from scipy.spatial import distance
from sklearn.neighbors import NearestNeighbors


class DummyEstimator:
    def __init__(self, cate_estimate: np.ndarray,
                 effect_intervals: Optional[np.ndarray] = None):
        self.cate_estimate = cate_estimate
        self.effect_intervals = effect_intervals

    def const_marginal_effect(self, X):
        return self.cate_estimate


def supported_metrics(
        problem: str,
        multivalue: bool,
        scores_only: bool) -> List[str]:
    if problem == "iv":
        metrics = ["energy_distance"]
        if not scores_only:
            metrics.append("ate")
        return metrics
    elif problem == "backdoor":
        # print("backdoor")
        if multivalue:
            # TODO: support other metrics for the multivalue case
            return ["energy_distance", "psw_energy_distance"]
        else:
            metrics = [
                "erupt",
                "norm_erupt",
                "prob_erupt",  # NEW
                "policy_risk",  # NEW
                "qini",
                "auc",
                # "r_scorer",
                "energy_distance",
                "psw_energy_distance",
                "frobenius_norm",  # NEW
                "codec"  # NEW
            ]
            if not scores_only:
                metrics.append("ate")
            return metrics


class Scorer:
    def __init__(
        self,
        causal_model: CausalModel,
        propensity_model: Any,
        problem: str,
        multivalue: bool,
    ):
        """
        Contains scoring logic for CausalTune.

        Access methods and attributes via `CausalTune.scorer`.

        """

        self.problem = problem
        self.multivalue = multivalue
        self.causal_model = copy.deepcopy(causal_model)

        self.identified_estimand = causal_model.identify_effect(
            proceed_when_unidentifiable=True
        )

        if problem == "backdoor":
            print(
                "Fitting a Propensity-Weighted scoring estimator "
                "to be used in scoring tasks"
            )
            treatment_series = causal_model._data[causal_model._treatment[0]]
            # this will also fit self.propensity_model, which we'll also use in
            # self.erupt
            self.psw_estimator = self.causal_model.estimate_effect(
                self.identified_estimand,
                method_name="backdoor.causaltune.models.MultivaluePSW",
                control_value=0,
                treatment_value=treatment_values(treatment_series, 0),
                target_units="ate",  # condition used for CATE
                confidence_intervals=False,
                method_params={
                    "init_params": {"propensity_model": propensity_model},
                },
            ).estimator

            if not hasattr(
                    self.psw_estimator,
                    'estimator') or not hasattr(
                    self.psw_estimator.estimator,
                    'propensity_model'):
                raise ValueError(
                    "Propensity model fitting failed. Please check the setup.")
            else:
                print("Propensity Model Fitted Successfully")

            treatment_name = self.psw_estimator._treatment_name
            if not isinstance(treatment_name, str):
                treatment_name = treatment_name[0]

            # No need to call self.erupt.fit()
            # as propensity model is already fitted
            # self.propensity_model = est.propensity_model
            self.erupt = ERUPT(
                treatment_name=treatment_name,
                propensity_model=self.psw_estimator.estimator.propensity_model,
                X_names=self.psw_estimator._effect_modifier_names
                + self.psw_estimator._observed_common_causes_names,
            )

    def ate(self, df: pd.DataFrame) -> tuple:
        """
        Calculate the Average Treatment Effect. Provide naive std estimates in
        single-treatment cases.

        Args:
            df (pandas.DataFrame): input dataframe

        Returns:
            tuple: tuple containing the ATE, standard deviation of the estimate
            (or None if multi-treatment), and sample size (or None if estimate
            has more than one dimension)
        """

        estimate = self.psw_estimator.estimator.effect(df).mean(axis=0)

        if len(estimate) == 1:
            # for now, let's cheat on the std estimation, take that from the
            # naive ate
            treatment_name = self.causal_model._treatment[0]
            outcome_name = self.causal_model._outcome[0]
            naive_est = Scorer.naive_ate(df[treatment_name], df[outcome_name])
            return estimate[0], naive_est[1], naive_est[2]
        else:
            return estimate, None, None

    def resolve_metric(self, metric: str) -> str:
        """Check if supplied metric is supported.
            If not, default to 'energy_distance'.

        Args:
            metric (str): evaluation metric

        Returns:
            str: metric/'energy_distance'

        """

        metrics = supported_metrics(
            self.problem, self.multivalue, scores_only=True)

        if metric not in metrics:
            logging.warning(
                f"Using energy_distance metric as {metric} is not in the list "
                f"of supported metrics for this usecase ({str(metrics)})"
            )
            return "energy_distance"
        else:
            return metric

    def resolve_reported_metrics(
        self, metrics_to_report: Union[List[str], None], scoring_metric: str
    ) -> List[str]:
        """
        Check if supplied reporting metrics are valid.

        Args:
            metrics_to_report (Union[List[str], None]): list of strings
            specifying the evaluation metrics to compute. Possible options
            include 'ate', 'erupt', 'norm_erupt', 'qini', 'auc',
            'energy_distance', and 'psw_energy_distance'.
            scoring_metric (str): specified metric.

        Returns:
            List[str]: list of valid metrics.
        """

        metrics = supported_metrics(
            self.problem,
            self.multivalue,
            scores_only=False)

        if metrics_to_report is None:
            return metrics
        else:
            metrics_to_report = sorted(
                list(set(metrics_to_report + [scoring_metric])))
            for m in metrics_to_report.copy():
                if m not in metrics:
                    logging.warning(
                        f"Dropping the metric {m} for problem: {self.problem} \
                        : must be one of {metrics}"
                    )
                    metrics_to_report.remove(m)
        return metrics_to_report

    @staticmethod
    def energy_distance_score(
        estimate: CausalEstimate,
        df: pd.DataFrame,
    ) -> float:
        """
        Calculate energy distance score between treated and controls. For
        theoretical details, see Ramos-Carreño and Torrecilla (2023).

        Args:
            estimate (dowhy.causal_estimator.CausalEstimate): causal estimate
            to evaluate
            df (pandas.DataFrame): input dataframe

        Returns:
            float: energy distance score
        """

        Y0X, _, split_test_by = Scorer._Y0_X_potential_outcomes(estimate, df)

        YX_1 = Y0X[Y0X[split_test_by] == 1]
        YX_0 = Y0X[Y0X[split_test_by] == 0]
        select_cols = estimate.estimator._effect_modifier_names + ["yhat"]

        energy_distance_score = dcor.energy_distance(
            YX_1[select_cols], YX_0[select_cols]
        )

        return energy_distance_score

    @staticmethod
    def _Y0_X_potential_outcomes(estimate: CausalEstimate, df: pd.DataFrame):
        est = estimate.estimator
        # assert est.identifier_method in ["iv", "backdoor"]
        treatment_name = (
            est._treatment_name
            if isinstance(est._treatment_name, str)
            else est._treatment_name[0]
        )
        df["dy"] = estimate.estimator.effect_tt(df)
        df["yhat"] = df[est._outcome_name] - df["dy"]

        split_test_by = (
            est.estimating_instrument_names[0]
            if est.identifier_method == "iv"
            else treatment_name
        )
        Y0X = copy.deepcopy(df)

        return Y0X, treatment_name, split_test_by

    # NEW:
    def frobenius_norm_score(
        self,
        estimate: CausalEstimate,
        df: pd.DataFrame,
        sd_threshold: float = 1e-2,
    ) -> float:
        """
        Calculate Frobenius norm-based score between treated and controls,
        using propensity score weighting.

        Args:
            estimate (CausalEstimate): causal estimate to evaluate
            df (pandas.DataFrame): input dataframe
            sd_threshold (float): threshold for standard deviation of CATE
            estimates

        Returns:
            float: Frobenius norm-based score, or np.inf if calculation is
            not possible
        """
        # Attempt to get CATE estimates, handling potential AttributeErrors
        try:
            cate_estimates = estimate.estimator.effect(df)
        except AttributeError:
            try:
                cate_estimates = estimate.estimator.effect_tt(df)
            except AttributeError:
                return np.inf  # Return inf if neither method is available

        # Check if CATE estimates are consistently constant (below threshold)
        if np.std(cate_estimates) <= sd_threshold:
            return np.inf  # Return inf for constant CATE estimates

        # Prepare data for treated and control groups
        Y0X, treatment_name, split_test_by = self._Y0_X_potential_outcomes(
            estimate, df)
        Y0X_1 = Y0X[Y0X[split_test_by] == 1]  # Treated group
        Y0X_0 = Y0X[Y0X[split_test_by] == 0]  # Control group

        # Check if either group is empty
        if len(Y0X_1) == 0 or len(Y0X_0) == 0:
            return np.inf  # Return inf if either group is empty

        # Select columns for analysis
        select_cols = estimate.estimator._effect_modifier_names + ["yhat"]

        # Calculate propensity scores for treated group
        propensitymodel = self.psw_estimator.estimator.propensity_model
        YX_1_all_psw = propensitymodel.predict_proba(
            Y0X_1[
                self.causal_model.get_effect_modifiers()
                + self.causal_model.get_common_causes()
            ]
        )
        treatment_series = Y0X_1[treatment_name]
        YX_1_psw = np.zeros(YX_1_all_psw.shape[0])
        for i in treatment_series.unique():
            YX_1_psw[treatment_series == i] = (
                YX_1_all_psw[:, i][treatment_series == i]
            )

        # Calculate propensity scores for control group
        propensitymodel = self.psw_estimator.estimator.propensity_model
        YX_0_psw = propensitymodel.predict_proba(
            Y0X_0[
                self.causal_model.get_effect_modifiers()
                + self.causal_model.get_common_causes()
            ]
        )[:, 0]

        # Ensure both datasets have the same number of rows
        min_rows = min(len(Y0X_1), len(Y0X_0))
        Y0X_1 = Y0X_1.iloc[:min_rows]
        Y0X_0 = Y0X_0.iloc[:min_rows]
        YX_1_psw = YX_1_psw[:min_rows]
        YX_0_psw = YX_0_psw[:min_rows]

        # Calculate the difference matrix with propensity score weights
        D = (Y0X_1[select_cols].values - Y0X_0[select_cols].values) * \
            np.sqrt(YX_1_psw * YX_0_psw).reshape(-1, 1)

        # Compute Frobenius norm of the weighted difference matrix
        frobenius_norm = np.linalg.norm(D, ord='fro')

        # Normalize the Frobenius norm by sqrt(n * p) where n is number of
        # samples and p is number of features
        n, p = D.shape
        normalized_score = frobenius_norm / np.sqrt(n * p)

        # Return the normalized score if it's finite, otherwise return infinity
        return normalized_score if np.isfinite(normalized_score) else np.inf

    def psw_energy_distance(
        self,
        estimate: CausalEstimate,
        df: pd.DataFrame,
        normalise_features=False,
    ) -> float:
        """
        Calculate propensity score adjusted energy distance score between
        treated and controls.

        Features are normalized using the
        `sklearn.preprocessing.QuantileTransformer`.

        For theoretical details, see Ramos-Carreño and Torrecilla (2023).

        Args:
            estimate (dowhy.causal_estimator.CausalEstimate): causal estimate
            to evaluate.
            df (pandas.DataFrame): input dataframe.
            normalise_features (bool): whether to normalize features with
            `QuantileTransformer`.

        Returns:
            float: propensity-score weighted energy distance score.
        """

        Y0X, treatment_name, split_test_by = Scorer._Y0_X_potential_outcomes(
            estimate, df
        )

        Y0X_1 = Y0X[Y0X[split_test_by] == 1]
        Y0X_0 = Y0X[Y0X[split_test_by] == 0]

        propensitymodel = self.psw_estimator.estimator.propensity_model
        YX_1_all_psw = propensitymodel.predict_proba(
            Y0X_1[
                self.causal_model.get_effect_modifiers()
                + self.causal_model.get_common_causes()
            ]
        )
        treatment_series = Y0X_1[treatment_name]

        YX_1_psw = np.zeros(YX_1_all_psw.shape[0])
        for i in treatment_series.unique():
            YX_1_psw[treatment_series == i] = (
                YX_1_all_psw[:, i][treatment_series == i]
            )

        propensitymodel = self.psw_estimator.estimator.propensity_model
        YX_0_psw = propensitymodel.predict_proba(
            Y0X_0[
                self.causal_model.get_effect_modifiers()
                + self.causal_model.get_common_causes()
            ]
        )[:, 0]

        select_cols = estimate.estimator._effect_modifier_names + ["yhat"]
        features = estimate.estimator._effect_modifier_names

        xy_psw = psw_joint_weights(YX_1_psw, YX_0_psw)
        xx_psw = psw_joint_weights(YX_0_psw)
        yy_psw = psw_joint_weights(YX_1_psw)

        xy_mean_weights = np.mean(xy_psw)
        xx_mean_weights = np.mean(xx_psw)
        yy_mean_weights = np.mean(yy_psw)

        if normalise_features:
            qt = QuantileTransformer(n_quantiles=200)
            X_quantiles = qt.fit_transform(Y0X[features])

            Y0X_transformed = pd.DataFrame(
                X_quantiles, columns=features, index=Y0X.index
            )
            Y0X_transformed.loc[:, ["yhat", split_test_by]] = Y0X[
                ["yhat", split_test_by]
            ]

            Y0X_1 = Y0X_transformed[Y0X_transformed[split_test_by] == 1]
            Y0X_0 = Y0X_transformed[Y0X_transformed[split_test_by] == 0]

        exponent = 1
        distance_xy = np.reciprocal(xy_mean_weights) * np.multiply(
            xy_psw,
            dcor.distances.pairwise_distances(
                Y0X_1[select_cols], Y0X_0[select_cols], exponent=exponent
            ),
        )
        distance_yy = np.reciprocal(yy_mean_weights) * np.multiply(
            yy_psw, dcor.distances.pairwise_distances(
                Y0X_1[select_cols], exponent=exponent), )
        distance_xx = np.reciprocal(xx_mean_weights) * np.multiply(
            xx_psw, dcor.distances.pairwise_distances(
                Y0X_0[select_cols], exponent=exponent), )
        psw_energy_distance = (
            2
            * np.mean(distance_xy)
            - np.mean(distance_xx)
            - np.mean(distance_yy))
        return psw_energy_distance

    # NEW:
    @staticmethod
    def default_policy(cate: np.ndarray) -> np.ndarray:
        """Default policy that assigns treatment if CATE > 0."""
        return (cate > 0).astype(int)

    # NEW:
    def policy_risk_score(
        self,
        estimate: CausalEstimate,
        df: pd.DataFrame,
        cate_estimate: np.ndarray,
        outcome_name: str,
        policy: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        rct_indices: Optional[pd.Index] = None,
        sd_threshold: float = 1e-2,
        clip: float = 0.05
    ) -> float:
        # Use default_policy if no custom policy is provided
        if policy is None:
            policy = self.default_policy

        # If no specific RCT indices are provided, use all indices
        if rct_indices is None:
            rct_indices = df.index

        # Ensure cate_estimate is a 1D array for consistent processing
        cate_estimate = np.squeeze(cate_estimate)

        # Return 0 if CATE estimates are consistently constant (below
        # threshold)
        if np.std(cate_estimate) <= sd_threshold:
            return 0  # This indicates no heterogeneity in treatment effects

        # Apply the policy to get treatment assignments based on CATE estimates
        policy_treatment = policy(cate_estimate)

        # Validate that the propensity model is properly fitted
        if not hasattr(
                self.psw_estimator,
                'estimator') or not hasattr(
                self.psw_estimator.estimator,
                'propensity_model'):
            raise ValueError(
                "Propensity model fitting failed. Please check the setup.")
        else:
            # Calculate propensity scores using the pre-fitted propensity model
            propensity_scores = (
                self.psw_estimator.estimator.propensity_model.predict_proba(
                    df.drop(
                        [
                            'index',
                            'variant',
                            'Y',
                            'dy',
                            'yhat'
                        ],
                        axis=1
                    )
                )
            )
            if propensity_scores.ndim == 2:
                # Use second column if 2D array
                propensity_scores = propensity_scores[:, 1]

            # Clip propensity scores to avoid extreme weights
            propensity_scores = np.clip(propensity_scores, clip, 1 - clip)

        treatment_name = self.psw_estimator._treatment_name

        # Calculate inverse probability weights
        weights = np.where(df[treatment_name] == 1,
                           1 / propensity_scores,
                           1 / (1 - propensity_scores))

        # Prepare RCT subset for analysis
        rct_df = df.loc[rct_indices].copy()
        rct_df['weight'] = weights[rct_indices]
        rct_df['policy_treatment'] = policy_treatment[rct_indices]

        # Compute policy value using inverse probability weighting
        value_policy = (
            (
                (rct_df[outcome_name] * (rct_df[treatment_name] == 1)
                 * (rct_df['policy_treatment'] == 1)
                 * rct_df['weight']).sum()
                / rct_df['weight'].sum()
                * (rct_df['policy_treatment'] == 1).mean()
            ) + (
                (rct_df[outcome_name] * (rct_df[treatment_name] == 0)
                 * (rct_df['policy_treatment'] == 0)
                 * rct_df['weight']).sum()
                / rct_df['weight'].sum()
                * (rct_df['policy_treatment'] == 0).mean()
            )
        )

        # Compute Policy Risk (1 - policy value)
        policy_risk = 1 - value_policy

        return policy_risk

    @staticmethod
    def qini_make_score(
        estimate: CausalEstimate, df: pd.DataFrame, cate_estimate: np.ndarray
    ) -> float:
        """
        Calculate the Qini score, defined as the area between the Qini curves
        of a model and random.

        Args:
            estimate (dowhy.causal_estimator.CausalEstimate): causal estimate
            to evaluate
            df (pandas.DataFrame): input dataframe
            cate_estimate (np.ndarray): array with CATE estimates

        Returns:
            float: Qini score
        """

        est = estimate.estimator
        new_df = pd.DataFrame()
        new_df["y"] = df[est._outcome_name]
        treatment_name = est._treatment_name
        if not isinstance(treatment_name, str):
            treatment_name = treatment_name[0]
        new_df["w"] = df[treatment_name]
        new_df["model"] = cate_estimate

        qini_score = metrics.qini_score(new_df)

        return qini_score["model"]

    # NEW
    @staticmethod
    def randomNN(ids):
        """
        Generate a list of random nearest neighbors.

        Parameters:
        ids (array-like): List of indices to sample from.

        Returns:
        numpy.ndarray: Array of sampled indices with
        no position i having x[i] == i.
        """

        m = len(ids)
        # Sample random integers from 0 to m-2, size m, with replacement
        x = np.random.choice(m - 1, m, replace=True)
        # Adjust x to ensure no position i has x[i] == i
        x = x + (x >= np.arange(m))
        return np.array(ids)[x]

    # NEW
    @staticmethod
    def estimateConditionalQ(Y, X, Z):
        """
        Estimate Q(Y, Z | X), the numerator of the measure of
        conditional dependence of Y on Z given X.

        Parameters:
        Y (array-like): Vector of responses (length n).
        X (array-like): Matrix of predictors (n by p).
        Z (array-like): Matrix of predictors (n by q).

        Returns:
        float: Estimation of Q(Y, Z | X).
        """

        # Ensure X and Z are numpy arrays
        if not isinstance(X, np.ndarray):
            X = np.array(X)
        if not isinstance(Z, np.ndarray):
            Z = np.array(Z)

        # To turn Z from shape (n,) to (n,1)
        Z = Z.reshape(-1, 1)

        n = len(Y)
        W = np.hstack((X, Z))

        # Compute the nearest neighbor of X
        nn_X = NearestNeighbors(n_neighbors=3, algorithm='auto').fit(X)
        nn_dists_X, nn_indices_X = nn_X.kneighbors(X)
        nn_index_X = nn_indices_X[:, 1]

        # Handle repeated data
        repeat_data = np.where(nn_dists_X[:, 1] == 0)[0]
        df_X = pd.DataFrame(
            {'id': repeat_data, 'group': nn_indices_X[repeat_data, 0]})
        df_X['rnn'] = df_X.groupby('group')['id'].transform(Scorer.randomNN)
        nn_index_X[repeat_data] = df_X['rnn'].values

        # Nearest neighbors with ties
        ties = np.where(nn_dists_X[:, 1] == nn_dists_X[:, 2])[0]
        ties = np.setdiff1d(ties, repeat_data)

        if len(ties) > 0:
            def helper_ties(a):
                distances = distance.cdist(X[a].reshape(
                    1, -1), np.delete(X, a, axis=0)).flatten()
                ids = np.where(distances == distances.min())[0]
                x = np.random.choice(ids)
                return x + (x >= a)

            nn_index_X[ties] = [helper_ties(a) for a in ties]

        # Compute the nearest neighbor of W
        nn_W = NearestNeighbors(n_neighbors=3, algorithm='auto').fit(W)
        nn_dists_W, nn_indices_W = nn_W.kneighbors(W)
        nn_index_W = nn_indices_W[:, 1]

        repeat_data = np.where(nn_dists_W[:, 1] == 0)[0]
        df_W = pd.DataFrame(
            {'id': repeat_data, 'group': nn_indices_W[repeat_data, 0]})
        df_W['rnn'] = df_W.groupby('group')['id'].transform(Scorer.randomNN)
        nn_index_W[repeat_data] = df_W['rnn'].values

        # Nearest neighbors with ties
        ties = np.where(nn_dists_W[:, 1] == nn_dists_W[:, 2])[0]
        ties = np.setdiff1d(ties, repeat_data)

        if len(ties) > 0:
            nn_index_W[ties] = [helper_ties(a) for a in ties]

        # Estimate Q
        R_Y = np.argsort(np.argsort(Y))  # Rank Y with ties method 'max'
        Q_n = (np.sum(np.minimum(R_Y, R_Y[nn_index_W]))
               - np.sum(np.minimum(R_Y, R_Y[nn_index_X]))) / (n**2)

        return Q_n

    # NEW
    @staticmethod
    def estimateConditionalS(Y, X):
        """
        Estimate S(Y, X), the denominator of the
        measure of dependence of Y on Z given X.

        Parameters:
        Y (array-like): Vector of responses (length n).
        X (array-like): Matrix of predictors (n by p).

        Returns:
        float: Estimation of S(Y, X).
        """

        # Ensure X is a numpy array
        if not isinstance(X, np.ndarray):
            X = np.array(X)

        n = len(Y)

        # Compute the nearest neighbor of X
        nn_X = NearestNeighbors(n_neighbors=3, algorithm='auto').fit(X)
        nn_dists_X, nn_indices_X = nn_X.kneighbors(X)
        nn_index_X = nn_indices_X[:, 1]

        # Handle repeated data
        repeat_data = np.where(nn_dists_X[:, 1] == 0)[0]
        df_X = pd.DataFrame(
            {'id': repeat_data, 'group': nn_indices_X[repeat_data, 0]})
        df_X['rnn'] = df_X.groupby('group')['id'].transform(Scorer.randomNN)
        nn_index_X[repeat_data] = df_X['rnn'].values

        # Nearest neighbors with ties
        ties = np.where(nn_dists_X[:, 1] == nn_dists_X[:, 2])[0]
        ties = np.setdiff1d(ties, repeat_data)

        if len(ties) > 0:
            def helper_ties(a):
                distances = distance.cdist(X[a].reshape(
                    1, -1), np.delete(X, a, axis=0)).flatten()
                ids = np.where(distances == distances.min())[0]
                x = np.random.choice(ids)
                return x + (x >= a)

            nn_index_X[ties] = [helper_ties(a) for a in ties]

        # Estimate S
        R_Y = np.argsort(np.argsort(Y))  # Rank Y with ties method 'max'
        S_n = np.sum(R_Y - np.minimum(R_Y, R_Y[nn_index_X])) / (n**2)

        return S_n

    # NEW
    @staticmethod
    def estimateConditionalT(Y, Z, X):
        """
        Estimate T(Y, Z | X), the measure of dependence of Y on Z given X.

        Parameters:
        Y (array-like): Vector of responses (length n).
        Z (array-like): Matrix of predictors (n by q).
        X (array-like): Matrix of predictors (n by p).

        Returns:
        float: Estimation of T(Y, Z | X).
        """

        S = Scorer.estimateConditionalS(Y, X)

        # Happens only if Y is constant
        if S == 0:
            return 1
        else:
            return Scorer.estimateConditionalQ(Y, X, Z) / S

    # NEW
    @staticmethod
    def codec(Y, Z, X=None, na_rm=True):
        """
        Estimate the conditional dependence coefficient (CODEC).

        The conditional dependence coefficient (CODEC) is a measure of the
        amount of conditional dependence between a random variable Y and a
        random vector Z given a random vector X, based on an i.i.d. sample of
        (Y, Z, X). The coefficient is asymptotically guaranteed to be between
        0 and 1.

        Parameters:
            Y (array-like): Vector of responses (length n).
            Z (array-like): Matrix of predictors (n by q).
            X (array-like, optional): Matrix of predictors (n by p). Default
            is None.
            na_rm (bool): If True, remove NAs.

        Returns:
            float: The conditional dependence coefficient (CODEC) of Y and Z
            given X. If X is None, this is just a measure of the dependence
            between Y and Z.

        References:
            Azadkia, M. and Chatterjee, S. (2019). A simple measure of
            conditional dependence. https://arxiv.org/pdf/1910.12327.pdf
        """

        if X is None:
            # Ensure inputs are in proper format
            if not isinstance(Y, np.ndarray):
                Y = np.array(Y)
            if not isinstance(Z, np.ndarray):
                Z = np.array(Z)
                # print(f"Shape of Z: {Z.shape}")
                # print(f"Z is: {Z}")

            if len(Y) != Z.shape[0]:
                raise ValueError("Number of rows of Y and Z should be equal.")
            if na_rm:
                # Remove NAs
                mask = np.isfinite(Y) & np.all(np.isfinite(Z), axis=1)
                Z = Z[mask]
                Y = Y[mask]

            n = len(Y)
            if n < 2:
                raise ValueError(
                    "Number of rows with no NAs should be greater than 1.")

            return Scorer.estimateConditionalQ(Y, Z, np.zeros((n, 0)))

        # Ensure inputs are in proper format
        if not isinstance(Y, np.ndarray):
            Y = np.array(Y)
        if not isinstance(X, np.ndarray):
            X = np.array(X)
        if not isinstance(Z, np.ndarray):
            Z = np.array(Z)
        if len(Y) != X.shape[0] or len(
                Y) != Z.shape[0] or X.shape[0] != Z.shape[0]:
            raise ValueError("Number of rows of Y, X, and Z should be equal.")

        n = len(Y)
        if n < 2:
            raise ValueError(
                "Number of rows with no NAs should be greater than 1.")

        return Scorer.estimateConditionalT(Y, Z, X)

    # NEW
    @staticmethod
    def identify_confounders(
            df: pd.DataFrame,
            treatment_col: str,
            outcome_col: str) -> list:
        """
        Identify confounders in a DataFrame.

        Args:
            df (pd.DataFrame): Input dataframe
            treatment_col (str): Name of the treatment column
            outcome_col (str): Name of the outcome column

        Returns:
            list: List of confounders' column names
        """

        confounders = [
            col for col in df.columns if col not in [
                treatment_col,
                outcome_col,
                "random",
                "index"]]
        return confounders

    # NEW
    @staticmethod
    def codec_score(estimate: CausalEstimate, df: pd.DataFrame) -> float:
        """Calculate the CODEC score for the effect of treatment on y_factual.

        Args:
            estimate (CausalEstimate): causal estimate to evaluate
            df (pd.DataFrame): input dataframe

        Returns:
            float: CODEC score
        """
        est = estimate.estimator
        treatment_name = est._treatment_name if isinstance(
            est._treatment_name, str) else est._treatment_name[0]
        outcome_name = est._outcome_name
        confounders = Scorer.identify_confounders(
            df, treatment_name, outcome_name)

        ########
        cate_est = est.effect(df)
        standard_deviations = np.std(cate_est)

        df["dy"] = est.effect_tt(df)

        df["yhat"] = df[est._outcome_name] - df["dy"]

        # have to use corrected y, not y factual to get the estimators
        # contribution in
        Y = df["yhat"]
        Z = df[treatment_name]
        X = df[confounders]

        if standard_deviations < 0.01:
            return np.inf

        return Scorer.codec(Y, Z, X)

    @staticmethod
    def auc_make_score(
        estimate: CausalEstimate, df: pd.DataFrame, cate_estimate: np.ndarray
    ) -> float:
        """Calculate the area under the uplift curve.

        Args:
            estimate (dowhy.causal_estimator.CausalEstimate): causal estimate
            to evaluate
            df (pandas.DataFrame): input dataframe
            cate_estimate (np.ndarray): array with cate estimates

        Returns:
            float: area under the uplift curve

        """

        est = estimate.estimator
        new_df = pd.DataFrame()
        new_df["y"] = df[est._outcome_name]
        treatment_name = est._treatment_name
        if not isinstance(treatment_name, str):
            treatment_name = treatment_name[0]
        new_df["w"] = df[treatment_name]
        new_df["model"] = cate_estimate

        auc_score = metrics.auuc_score(new_df)

        return auc_score["model"]

    @staticmethod
    def real_qini_make_score(
        estimate: CausalEstimate, df: pd.DataFrame, cate_estimate: np.ndarray
    ) -> float:
        # TODO  To calculate the 'real' qini score for synthetic datasets, to
        # be done

        # est = estimate.estimator
        new_df = pd.DataFrame()

        # new_df['tau'] = [df['y_factual'] - df['y_cfactual']]
        new_df["model"] = cate_estimate

        qini_score = metrics.qini_score(new_df)

        return qini_score["model"]

    @staticmethod
    def r_make_score(
            estimate: CausalEstimate,
            df: pd.DataFrame,
            cate_estimate: np.ndarray,
            r_scorer) -> float:
        """
        Calculate r_score.

        For details, refer to Nie and Wager (2017) and Schuler et al. (2018).
        Adapted from the EconML implementation.

        Args:
            estimate (dowhy.causal_estimator.CausalEstimate): causal estimate
            to evaluate
            df (pandas.DataFrame): input dataframe
            cate_estimate (np.ndarray): array with CATE estimates
            r_scorer: callable object used to compute the R-score

        Returns:
            float: r_score
        """

        # TODO
        return r_scorer.score(cate_estimate)

    @staticmethod
    def naive_ate(treatment: pd.Series, outcome: pd.Series):
        """Calculate simple ATE.

        Args:
            treatment (pandas.Series): series of treatments
            outcome (pandas.Series): series of outcomes

        Returns:
            tuple: tuple of simple ATE, standard deviation, and sample size

        """

        treated = (treatment == 1).sum()

        mean_ = outcome[treatment == 1].mean() - outcome[treatment == 0].mean()
        std1 = outcome[treatment == 1].std() / (math.sqrt(treated) + 1e-3)
        std2 = outcome[treatment == 0].std() / (
            math.sqrt(len(outcome) - treated) + 1e-3
        )
        std_ = math.sqrt(std1 * std1 + std2 * std2)
        return (mean_, std_, len(treatment))

    def group_ate(
        self, df: pd.DataFrame, policy: Union[pd.DataFrame, np.ndarray]
    ) -> pd.DataFrame:
        """
        Compute the average treatment effect (ATE) for different groups
        specified by a policy.

        Args:
            df (pandas.DataFrame): input dataframe, should contain columns
            for the treatment, outcome, and policy.
            policy (Union[pd.DataFrame, np.ndarray]): policy column in df or
            an array of the policy values, used to group the data.

        Returns:
            pandas.DataFrame: ATE, std, and size per policy.
        """

        tmp = {"all": self.ate(df)}
        for p in sorted(list(policy.unique())):
            tmp[p] = self.ate(df[policy == p])

        tmp2 = [
            {"policy": str(p), "mean": m, "std": s, "count": c}
            for p, (m, s, c) in tmp.items()
        ]

        return pd.DataFrame(tmp2)

    def make_scores(
        self,
        estimate: CausalEstimate,
        df: pd.DataFrame,
        metrics_to_report: List[str],
        r_scorer=None,
    ) -> dict:
        """
        Calculate various performance metrics for a given causal estimate using
        a given DataFrame.

        Args:
            estimate (dowhy.causal_estimator.CausalEstimate): causal estimate
            to evaluate.
            df (pandas.DataFrame): input dataframe.
            metrics_to_report (List[str]): list of strings specifying the
            evaluation metrics to compute. Possible options include 'ate',
            'erupt', 'norm_erupt', 'qini', 'auc', 'energy_distance' and
            'psw_energy_distance'.
            r_scorer (Optional): callable object used to compute the R-score,
            default is None.

        Returns:
            dict: dictionary containing the evaluation metrics specified in
            metrics_to_report. The values key in the dictionary contains the
            input DataFrame with additional columns for the propensity scores,
            the policy, the normalized policy, and the weights, if applicable.
        """

        out = dict()
        df = df.copy().reset_index()

        est = estimate.estimator
        treatment_name = est._treatment_name
        if not isinstance(treatment_name, str):
            treatment_name = treatment_name[0]
        outcome_name = est._outcome_name

        cate_estimate = est.effect(df)

        # TODO: fix this hack with proper treatment of multivalues
        if len(cate_estimate.shape) > 1 and cate_estimate.shape[1] == 1:
            cate_estimate = cate_estimate.reshape(-1)

        # TODO: fix this, currently broken
        # covariates = est._effect_modifier_names
        # Include CATE Interpereter for both IV and CATE models
        # intrp = SingleTreeCateInterpreter(
        #     include_model_uncertainty=False, max_depth=2, min_samples_leaf=10
        # )
        # intrp.interpret(DummyEstimator(cate_estimate), df[covariates])
        # intrp.feature_names = covariates
        # out["intrp"] = intrp

        if self.problem == "backdoor":
            values = df[[treatment_name, outcome_name]]
            simple_ate = self.ate(df)[0]

            if isinstance(simple_ate, float):
                # simple_ate = simple_ate[0]
                # .reset_index(drop=True)
                propensitymodel = self.psw_estimator.estimator.propensity_model
                values["p"] = (
                    propensitymodel.predict_proba(
                        df[
                            self.causal_model.get_effect_modifiers()
                            + self.causal_model.get_common_causes()
                        ]
                    )[:, 1]
                )
                values["policy"] = cate_estimate > 0
                values["norm_policy"] = cate_estimate > simple_ate
                values["weights"] = self.erupt.weights(
                    df, lambda x: cate_estimate > 0
                )
            else:
                pass
                # TODO: what do we do here if multiple treatments?

            if "erupt" in metrics_to_report:
                erupt_score = self.erupt.score(
                    df, df[outcome_name], cate_estimate > 0)
                out["erupt"] = erupt_score

            if "norm_erupt" in metrics_to_report:
                norm_erupt_score = (
                    self.erupt.score(
                        df,
                        df[outcome_name],
                        cate_estimate > simple_ate
                    ) - simple_ate * values["norm_policy"].mean()
                )
                out["norm_erupt"] = norm_erupt_score

            if "prob_erupt" in metrics_to_report:
                treatment_effects = pd.Series(cate_estimate, index=df.index)
                treatment_std_devs = pd.Series(
                    cate_estimate.std(), index=df.index)
                prob_erupt_score = self.erupt.probabilistic_erupt_score(
                    df, df[outcome_name],
                    treatment_effects,
                    treatment_std_devs
                )
                out["prob_erupt"] = prob_erupt_score

            if "frobenius_norm" in metrics_to_report:
                out["frobenius_norm"] = self.frobenius_norm_score(estimate, df)

            if "policy_risk" in metrics_to_report:
                try:
                    out["policy_risk"] = self.policy_risk_score(
                        estimate=estimate,
                        df=df,
                        cate_estimate=cate_estimate,
                        outcome_name=outcome_name,
                        policy=None
                    )
                except Exception as e:
                    print(e)
                    pass

            if "qini" in metrics_to_report:
                out["qini"] = Scorer.qini_make_score(
                    estimate, df, cate_estimate)

            if "auc" in metrics_to_report:
                out["auc"] = Scorer.auc_make_score(estimate, df, cate_estimate)

            if r_scorer is not None:
                out["r_score"] = Scorer.r_make_score(
                    estimate, df, cate_estimate, r_scorer
                )

            # values = values.rename(columns={treatment_name: "treated"})
            assert len(values) == len(
                df), "Index weirdness when adding columns!"
            values = values.copy()
            out["values"] = values

        if "ate" in metrics_to_report:
            out["ate"] = cate_estimate.mean()
            out["ate_std"] = cate_estimate.std()

        if "energy_distance" in metrics_to_report:
            out["energy_distance"] = Scorer.energy_distance_score(estimate, df)

        if "psw_energy_distance" in metrics_to_report:
            out["psw_energy_distance"] = self.psw_energy_distance(
                estimate,
                df,
            )
        if "codec" in metrics_to_report:
            temp = self.codec_score(estimate, df)
            out["codec"] = temp

        del df
        return out

    @staticmethod
    def best_score_by_estimator(
        scores: Dict[str, dict], metric: str
    ) -> Dict[str, dict]:
        """Obtain best score for each estimator.

        Args:
            scores (Dict[str, dict]): CausalTune.scores dictionary
            metric (str): metric of interest

        Returns:
            Dict[str, dict]: dictionary containing best score by estimator

        """

        for k, v in scores.items():
            if "estimator_name" not in v:
                raise ValueError(
                    f"Malformed scores dict, 'estimator_name' field missing "
                    f"in{k}, {v}"
                )

        estimator_names = sorted(
            list(
                set(
                    [
                        v["estimator_name"]
                        for v in scores.values()
                        if "estimator_name" in v
                    ]
                )
            )
        )
        best = {}
        for name in estimator_names:
            est_scores = [
                v
                for v in scores.values()
                if "estimator_name" in v and v["estimator_name"] == name
            ]
            best[name] = (
                min(
                    est_scores,
                    key=lambda x: x[metric]) if metric in [
                    "energy_distance",
                    "psw_energy_distance",
                    "frobenius_norm",
                    "codec",
                    "policy_risk"] else max(
                    est_scores,
                    key=lambda x: x[metric]))

        return best



import numpy as np
from flaml import tune
from copy import deepcopy
from typing import Optional, Sequence, Union, Iterable, Dict, Any, Tuple
from dataclasses import dataclass, field

import warnings
from econml.inference import BootstrapInference  # noqa F401
from sklearn import linear_model

from causaltune.utils import clean_config
from causaltune.search.component import model_from_cfg, joint_config


@dataclass
class EstimatorConfig:
    outcome_model_name: str = None
    final_model_name: str = None
    propensity_model_name: str = None
    init_params: dict = field(default_factory=dict)
    fit_params: dict = field(default_factory=dict)
    search_space: dict = field(default_factory=dict)
    defaults: dict = field(default_factory=dict)
    supports_multivalue: bool = False
    experimental: bool = False
    inference: str = "bootstrap"


class SimpleParamService:
    def __init__(
        self,
        multivalue: bool,
        n_bootstrap_samples: Optional[int] = None,
        n_jobs: Optional[int] = None,
        include_experimental=False,
        sample_outcome_estimators: bool = False,
    ):
        self.n_jobs = n_jobs
        self.include_experimental = include_experimental
        self.n_bootstrap_samples = n_bootstrap_samples
        self.multivalue = multivalue
        self.sample_outcome_estimators = sample_outcome_estimators

    def estimator_names_from_patterns(
        self,
        problem: str,
        patterns: Union[Sequence, str],
        data_rows: Optional[int] = None,
    ):
        def problem_match(est_name: str, problem: str) -> bool:
            return est_name.split(".")[0] == problem

        if patterns == "all":
            if data_rows <= 1000:
                return [e for e in self.estimator_names if problem_match(e, problem)]
            else:
                warnings.warn(
                    "Excluding OrthoForests as they can have problems with large datasets"
                )
                return [
                    e
                    for e in self.estimator_names
                    if ("OrthoForest" not in e) and (problem_match(e, problem))
                ]

        elif patterns == "cheap_inference":
            cfgs = self._configs()
            ests = [
                est
                for est in self.estimator_names
                if cfgs[est].inference != "bootstrap" and problem_match(est, problem)
            ]
            if data_rows <= 1000:
                return ests
            else:
                warnings.warn(
                    "Excluding OrthoForests as they can have problems with large datasets"
                )
                return [e for e in ests if ("OrthoForest" not in e)]

        elif patterns == "auto":
            if problem == "backdoor":
                if self.multivalue:
                    patterns = ["LinearDML"]
                else:
                    # These are the ones we've seen best results from, empirically,
                    # plus dummy for baseline, and SLearner as that's the simplest possible
                    patterns = [
                        "Dummy",
                        "NewDummy",
                        "SLearner",
                        "DomainAdaptationLearner",
                        "TransformedOutcome",
                        "CausalForestDML",
                        "ForestDRLearner",
                    ]

            elif problem == "iv":
                patterns = [
                    "DMLIV",
                    "LinearDRIV",
                    "OrthoIV",
                    "SparseLinearDRIV",
                    "LinearIntentToTreatDRIV",
                ]
            return self.estimator_names_from_patterns(problem, patterns)
        else:
            try:
                for p in patterns:
                    assert isinstance(p, str)
            except Exception:
                raise ValueError(
                    "Invalid estimator list, must be 'auto', 'all', 'cheap_inference' or a list of strings"
                )

            out = [
                est
                for p in patterns
                for est in self.estimator_names
                if p in est and problem_match(est, problem)
            ]
            return sorted(list(set(out)))

    @property
    def estimator_names(self):
        cfgs = self._configs()
        if self.multivalue:
            cfgs = {k: v for k, v in cfgs.items() if v.supports_multivalue}

        if self.include_experimental:
            return list(cfgs.keys())
        else:
            return [est for est, cfg in cfgs.items() if not cfg.experimental]

    def search_space(
        self,
        estimator_list: Iterable[str],
        data_size: Tuple[int, int],
        outcome_estimator_list: Iterable[str] = None,
    ):
        """Constructs search space with estimators and their respective configs

        Args:
            estimator_list (Iterable[str]): estimators to consider

        Returns:
            dict: hierarchical search space
        """
        search_space = [
            {
                "estimator_name": est,
                **est_params.search_space,
            }
            for est, est_params in self._configs().items()
            if est in estimator_list
        ]

        out = {"estimator": tune.choice(search_space)}
        if self.sample_outcome_estimators:
            out["outcome_estimator"], _, _ = joint_config(
                data_size, outcome_estimator_list
            )

        return out

    def default_configs(
        self,
        estimator_list: Iterable[str],
        data_size: Tuple[int, int],
        outcome_estimator_list: Iterable[str] = None,
        num_outcome_samples: int = 3,
    ):
        """Creates list with initial configs to try before moving
        on to hierarchical HPO.
        The list has been identified by evaluating performance of all
        learners on a range of datasets (and metrics).
        Each entry is a dictionary with a learner and its best-performing
        hyper params
        TODO: identify best_performers for list below

        Args:
            estimator_list (Iterable[str]): estimators to consider

        Returns:
            list: list of dicts with promising initial configs
        """
        pre_points = [
            {"estimator": {"estimator_name": est, **est_params.defaults}}
            for est, est_params in self._configs().items()
            if est in estimator_list
        ]

        cfgs = self._configs()

        if self.sample_outcome_estimators:
            points = []
            _, init_params, _ = joint_config(data_size, outcome_estimator_list)
            for p in pre_points:
                if cfgs[p["estimator"]["estimator_name"]].outcome_model_name is None:
                    this_p = deepcopy(p)
                    # this won't have any effect, so pick any valid config to mitigate sampling bias
                    this_p["outcome_estimator"] = np.random.choice(init_params)
                    points.append(p)
                    continue
                else:  # Sample different outcome functions for first pass
                    for outcome_est in np.random.choice(
                        init_params, size=num_outcome_samples, replace=False
                    ):
                        this_p = deepcopy(p)
                        this_p["outcome_estimator"] = outcome_est
                        points.append(this_p)
        else:
            points = pre_points

        return points

    def method_params(
        self,
        config: dict,
        outcome_model: Any,
        propensity_model: Any,
        final_model: Any = None,
    ):
        est_config = clean_config(deepcopy(config["estimator"]))
        estimator_name = est_config.pop("estimator_name")

        cfg = self._configs()[estimator_name]

        if outcome_model == "auto" and cfg.outcome_model_name is not None:
            # Spawn the outcome model dynamically
            outcome_model = model_from_cfg(config["outcome_estimator"])

        if (
            cfg.outcome_model_name is not None
            and cfg.outcome_model_name not in cfg.init_params
        ):
            cfg.init_params[cfg.outcome_model_name] = deepcopy(outcome_model)

        if (
            cfg.propensity_model_name is not None
            and cfg.propensity_model_name not in cfg.init_params
        ):
            cfg.init_params[cfg.propensity_model_name] = deepcopy(propensity_model)

        if (
            cfg.final_model_name is not None
            and cfg.final_model_name not in cfg.init_params
        ):
            cfg.init_params[cfg.final_model_name] = (
                deepcopy(final_model)
                if final_model is not None
                else deepcopy(outcome_model)
            )

        method_params = {
            "init_params": {**deepcopy(est_config), **cfg.init_params},
            "fit_params": {},
        }
        return method_params

    def full_config(self, estimator_name: str):
        cfg = self._configs()[estimator_name]
        return cfg

    def _configs(self) -> Dict[str, EstimatorConfig]:
        if self.n_bootstrap_samples is not None:
            # TODO Egor please look into this
            # bootstrap is causing recursion errors (see notes below)
            # bootstrap = BootstrapInference(
            #     n_bootstrap_samples=self.n_bootstrap_samples, n_jobs=self.n_jobs
            # )
            pass

        configs: dict[str:EstimatorConfig] = {
            "backdoor.causaltune.models.NaiveDummy": EstimatorConfig(),
            "backdoor.causaltune.models.Dummy": EstimatorConfig(
                propensity_model_name="propensity_model",
                experimental=False,
            ),
            "backdoor.propensity_score_weighting": EstimatorConfig(
                propensity_model_name="propensity_model",
                experimental=True,
            ),
            "backdoor.econml.metalearners.SLearner": EstimatorConfig(
                outcome_model_name="overall_model",
                supports_multivalue=True,
            ),
            "backdoor.econml.metalearners.TLearner": EstimatorConfig(
                outcome_model_name="models",
                supports_multivalue=True,
            ),
            "backdoor.econml.metalearners.XLearner": EstimatorConfig(
                outcome_model_name="models",
                propensity_model_name="propensity_model",
                supports_multivalue=True,
            ),
            "backdoor.econml.metalearners.DomainAdaptationLearner": EstimatorConfig(
                outcome_model_name="models",
                propensity_model_name="propensity_model",
                final_model_name="final_models",
                supports_multivalue=True,
            ),
            "backdoor.econml.dr.ForestDRLearner": EstimatorConfig(
                outcome_model_name="model_regression",
                propensity_model_name="model_propensity",
                init_params={
                    # putting these here for now, until default values can be reconciled with search space
                    "mc_iters": None,
                    "max_depth": None,
                },
                search_space={
                    "min_propensity": tune.loguniform(1e-6, 1e-1),
                    # "mc_iters": tune.randint(0, 10), # is this worth searching over?
                    "n_estimators": tune.randint(2, 200),
                    # "max_depth": is tune.choice([None, tune.randint(2, 1000)]) the right syntax?
                    "min_samples_split": tune.randint(2, 20),
                    "min_samples_leaf": tune.randint(1, 25),
                    "min_weight_fraction_leaf": tune.uniform(0, 0.5),
                    "max_features": tune.choice(["auto", "sqrt", "log2", None]),
                    "min_impurity_decrease": tune.uniform(0, 10),
                    "max_samples": tune.uniform(1e-6, 0.5),
                    "min_balancedness_tol": tune.uniform(0, 0.5),
                    "honest": tune.choice([0, 1]),
                    "subforest_size": tune.randint(2, 10),
                },
                defaults={
                    "min_propensity": 1e-6,
                    "n_estimators": 100,  # this is 1000 by default?
                    "min_samples_split": 5,
                    "min_samples_leaf": 5,
                    "min_weight_fraction_leaf": 0.0,
                    "max_features": "auto",
                    "min_impurity_decrease": 0.0,
                    "max_samples": 0.45,
                    "min_balancedness_tol": 0.45,
                    "honest": True,
                    "subforest_size": 4,
                },
                supports_multivalue=True,
                inference="blb",
            ),
            "backdoor.econml.dr.LinearDRLearner": EstimatorConfig(
                outcome_model_name="model_regression",
                propensity_model_name="model_propensity",
                init_params={
                    "mc_iters": None,
                },
                search_space={
                    "fit_cate_intercept": tune.choice([0, 1]),
                    "min_propensity": tune.loguniform(1e-6, 1e-1),
                    # "mc_iters": tune.randint(0, 10),
                },
                defaults={
                    "fit_cate_intercept": True,
                    "min_propensity": 1e-6,
                },
                supports_multivalue=True,
                inference="auto",
            ),
            "backdoor.econml.dr.SparseLinearDRLearner": EstimatorConfig(
                outcome_model_name="model_regression",
                propensity_model_name="model_propensity",
                init_params={
                    "mc_iters": None,
                },
                search_space={
                    "fit_cate_intercept": tune.choice([0, 1]),
                    "n_alphas": tune.lograndint(1, 1000),
                    "n_alphas_cov": tune.lograndint(1, 100),
                    "min_propensity": tune.loguniform(1e-6, 1e-1),
                    # "mc_iters": tune.randint(0, 10),
                    "tol": tune.qloguniform(1e-7, 1, 1e-7),
                    "max_iter": tune.qlograndint(100, 100000, 100),
                    "mc_agg": tune.choice(["mean", "median"]),
                },
                defaults={
                    "fit_cate_intercept": True,
                    "n_alphas": 100,
                    "n_alphas_cov": 10,
                    "min_propensity": 1e-6,
                    "tol": 0.0001,
                    "max_iter": 10000,
                    "mc_agg": "mean",
                },
                supports_multivalue=True,
                inference="auto",
            ),
            "backdoor.econml.dml.LinearDML": EstimatorConfig(
                outcome_model_name="model_y",
                propensity_model_name="model_t",
                init_params={
                    "discrete_treatment": True,
                    # it runs out of memory fast if the below is not set
                    "linear_first_stages": False,
                    "mc_iters": None,
                },
                search_space={
                    "fit_cate_intercept": tune.choice([0, 1]),
                    # "mc_iters": tune.randint(0, 10),
                    "mc_agg": tune.choice(["mean", "median"]),
                },
                defaults={
                    "fit_cate_intercept": True,
                    "mc_agg": "mean",
                },
                supports_multivalue=True,
                inference="statsmodels",
            ),
            "backdoor.econml.dml.SparseLinearDML": EstimatorConfig(
                outcome_model_name="model_y",
                propensity_model_name="model_t",
                init_params={
                    "discrete_treatment": True,
                    # it runs out of memory fast if the below is not set
                    "linear_first_stages": False,
                    "mc_iters": None,
                },
                search_space={
                    "fit_cate_intercept": tune.choice([0, 1]),
                    # "mc_iters": tune.randint(0, 10),
                    "n_alphas": tune.lograndint(1, 1000),
                    "n_alphas_cov": tune.lograndint(1, 100),
                    "tol": tune.qloguniform(1e-7, 1, 1e-7),
                    "max_iter": tune.qlograndint(100, 100000, 100),
                    "mc_agg": tune.choice(["mean", "median"]),
                },
                defaults={
                    "fit_cate_intercept": True,
                    "n_alphas": 100,
                    "n_alphas_cov": 10,
                    "tol": 0.0001,
                    "max_iter": 10000,
                    "mc_agg": "mean",
                },
                supports_multivalue=True,
                inference="auto",
            ),
            "backdoor.econml.dml.CausalForestDML": EstimatorConfig(
                outcome_model_name="model_y",
                propensity_model_name="model_t",
                init_params={
                    # "max_depth": self.max_depth,
                    # "n_estimators": self.n_estimators,
                    "discrete_treatment": True,
                    # "inference": False,
                    "mc_iters": None,
                    "max_depth": None,
                    "min_var_fraction_leaf": None,
                },
                search_space={
                    # "mc_iters": tune.randint(0, 10),
                    "drate": tune.choice([0, 1]),
                    "n_estimators": tune.randint(2, 500),
                    "criterion": tune.choice(["mse", "het"]),
                    # "max_depth": tune.randint(2, 1000),
                    "min_samples_split": tune.randint(2, 30),
                    "min_samples_leaf": tune.randint(1, 25),
                    "min_weight_fraction_leaf": tune.uniform(0, 0.5),
                    # "min_var_fraction_leaf": tune.uniform(0, 1),
                    "max_features": tune.choice(["auto", "sqrt", "log2", None]),
                    "min_impurity_decrease": tune.uniform(0, 10),
                    "max_samples": tune.uniform(1e-6, 0.5),
                    "min_balancedness_tol": tune.uniform(0, 0.5),
                    "honest": tune.choice([0, 1]),
                    # "inference": tune.choice([0, 1]),
                    "fit_intercept": tune.choice([0, 1]),
                    # Difficult as needs to be a factor of 'n_estimators'
                    "subforest_size": tune.randint(2, 10),
                },
                defaults={
                    # "mc_iters": tune.randint(0, 10),
                    "drate": True,
                    "n_estimators": 100,
                    "criterion": "mse",
                    "min_samples_split": 10,
                    "min_samples_leaf": 5,
                    "min_weight_fraction_leaf": 0.0,
                    "max_features": "auto",
                    "min_impurity_decrease": 0.0,
                    "max_samples": 0.45,
                    "min_balancedness_tol": 0.45,
                    "honest": True,
                    # "inference": tune.choice([0, 1]),
                    "fit_intercept": True,
                    "subforest_size": 4,
                },
                supports_multivalue=True,
                inference="auto",
            ),
            "backdoor.causaltune.models.TransformedOutcome": EstimatorConfig(
                outcome_model_name="outcome_model",
                propensity_model_name="propensity_model",
            ),
            # leaving out DML and NonParamDML as they're base classes for the 3
            # above
            #
            # This one breaks when running, need to figure out why
            # "backdoor.econml.dr.DRLearner": {
            #     "init_params": {
            #         "model_propensity": propensity_model,
            #         "model_regression": outcome_model,
            #         "model_final": final_model,
            #     },
            #     "fit_params": {},
            # },
            "backdoor.econml.orf.DROrthoForest": EstimatorConfig(
                propensity_model_name="propensity_model",
                init_params={
                    "model_Y": linear_model.Ridge(
                        alpha=0.01
                    ),  # WeightedLasso(alpha=0.01),  #
                    "n_jobs": self.n_jobs,
                    # "max_depth": self.max_depth,
                    # "n_trees": self.n_estimators,
                    # "min_leaf_size": self.min_leaf_size,
                    "backend": "loky",
                },
                search_space={
                    "n_trees": tune.randint(2, 750),
                    "min_leaf_size": tune.randint(1, 50),
                    "max_depth": tune.randint(2, 1000),
                    "subsample_ratio": tune.uniform(0, 1),
                    # "bootstrap": tune.choice([0, 1]),
                    "lambda_reg": tune.uniform(0, 1),
                },
                defaults={
                    "n_trees": 500,
                    "min_leaf_size": 10,
                    "max_depth": 10,
                    "subsample_ratio": 0.7,
                    "lambda_reg": 0.01,
                },
                experimental=True,  # OrthoForest estimators are notoriously slow
                supports_multivalue=True,
                inference="blb",
            ),
            "backdoor.econml.orf.DMLOrthoForest": EstimatorConfig(
                propensity_model_name="model_T",
                init_params={
                    "model_Y": linear_model.Ridge(
                        alpha=0.01
                    ),  # WeightedLasso(alpha=0.01),  #
                    "discrete_treatment": True,
                    "n_jobs": self.n_jobs,
                    # "max_depth": self.max_depth,
                    # "n_trees": self.n_estimators,
                    # "min_leaf_size": self.min_leaf_size,
                    # Loky was running out of disk space for some reason
                    "backend": "threading",
                },
                search_space={
                    "n_trees": tune.randint(2, 750),
                    "min_leaf_size": tune.randint(1, 50),
                    "max_depth": tune.randint(2, 1000),
                    "subsample_ratio": tune.uniform(0, 1),
                    # "bootstrap": tune.choice([0, 1]),
                    "lambda_reg": tune.uniform(0, 1),
                },
                defaults={
                    "n_trees": 500,
                    "min_leaf_size": 10,
                    "max_depth": 10,
                    "subsample_ratio": 0.7,
                    # "bootstrap": tune.choice([0, 1]),
                    "lambda_reg": 0.01,
                },
                experimental=True,  # OrthoForest estimators are notoriously slow
                supports_multivalue=True,
                inference="blb",
            ),
            "iv.econml.iv.dr.LinearDRIV": EstimatorConfig(
                outcome_model_name="model_y_xw",
                propensity_model_name="model_t_xw",
                search_space={
                    "projection": tune.choice([0, 1]),
                },
                defaults={"projection": True},
            ),
            "iv.econml.iv.dml.OrthoIV": EstimatorConfig(
                outcome_model_name="model_y_xw",
                propensity_model_name="model_t_xw",
                search_space={
                    "mc_agg": tune.choice(["mean", "median"]),
                },
                defaults={
                    "mc_agg": "mean",
                },
            ),
            "iv.econml.iv.dml.DMLIV": EstimatorConfig(
                outcome_model_name="model_y_xw",
                propensity_model_name="model_t_xw",
                search_space={
                    "mc_agg": tune.choice(["mean", "median"]),
                },
                defaults={
                    "mc_agg": "mean",
                },
            ),
            "iv.econml.iv.dr.SparseLinearDRIV": EstimatorConfig(
                outcome_model_name="model_y_xw",
                propensity_model_name="model_t_xw",
                search_space={
                    "projection": tune.choice([0, 1]),
                    "opt_reweighted": tune.choice([0, 1]),
                    "cov_clip": tune.quniform(0.08, 0.2, 0.01),
                },
                defaults={
                    "projection": 0,
                    "opt_reweighted": 0,
                    "cov_clip": 0.1,
                },
            ),
            "iv.econml.iv.dr.LinearIntentToTreatDRIV": EstimatorConfig(
                outcome_model_name="model_y_xw",
                search_space={
                    "cov_clip": tune.quniform(0.08, 0.2, 0.01),
                    "opt_reweighted": tune.choice([0, 1]),
                },
                defaults={
                    "cov_clip": 0.1,
                    "opt_reweighted": 1,
                },
            ),
        }
        return configs



from setuptools import find_packages, setup

with open("README.md") as f:
    long_description = f.read()

setup(
    name="causaltune",
    version="0.1.5",
    description="AutoML for Causal Inference.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Wise",
    url="https://github.com/py-why/causaltune",
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    install_requires=[
        "dowhy==0.9.1",
        "econml==0.14.1",
        "FLAML==2.2.0",
        "xgboost==1.7.6",
        "numpy==1.23.5",
        "pandas",
        "pytest",
        "scikit_learn",
        "matplotlib",
        "dcor",
        "holidays",
        "setuptools==65.5.1",
        "wise-pizza",
        "seaborn",
        "category_encoders==2.6.3",
    ],
    extras_require={
        "test": [
            "autoflake",
            "black==23.3.0",
            "flake8",
            "isort",
            "pytest",
            "pytest-cov",
            "nbmake",
        ],
        "ray": ["ray[tune]~=1.11.0"],
    },
    packages=find_packages(
        include=["causaltune", "causaltune.*"],
        exclude=["tests*"],
    ),
    include_package_data=True,
    keywords="causaltune",
)



import json
import numpy as np
import hashlib
import pandas as pd
from flaml import AutoML


# this hashing approach was inspired by  https://death.andgravity.com/stable-hashing


def get_hash(thing):
    return hashlib.md5(json_dumps(thing).encode("utf-8")).digest().hex()


def json_dumps(thing):
    return json.dumps(
        thing,
        default=json_default,
        ensure_ascii=False,
        sort_keys=True,
        indent=None,
        separators=(",", ":"),
    )


def json_default(thing):
    if isinstance(thing, np.ndarray):
        return str(bytes(thing))
    if isinstance(thing, pd.DataFrame):
        return (tuple(thing.columns), str(bytes(thing.values)))
    elif isinstance(thing, pd.Series):
        return str(bytes(thing.values))
    if isinstance(thing, type):
        return thing.__name__
    raise TypeError(f"object of type {type(thing).__name__} not serializable")


class Memoizer(dict):
    def __call__(self, fun, x):
        key = get_hash(x)
        if key in self:
            print("Reusing cached fit!")
        else:
            print("Running new fit!")
            self[key] = fun(**x)

        return self[key]


# why bother with a borg when a closure will do? :)
memoizer = Memoizer()


def fitter_fun(fitter: type, init_args, init_kwargs, fit_args, fit_kwargs):
    x = fitter(*init_args, **init_kwargs)
    print("calling fit method with ", fit_kwargs)
    x.fit(*fit_args, **fit_kwargs)
    return x


class MemoizingWrapper(AutoML):
    def __init__(self, *args, fit_params=None, **kwargs):
        self.init_args = args
        self.init_kwargs = kwargs
        if fit_params is not None:
            self.fit_params = fit_params
        else:
            self.fit_params = {}

    def fit(self, *args, **kwargs):
        # we defer the initialization to the fit() method so we can memoize the fit
        # using all the args from both init and fit
        used_kwargs = {**kwargs, **self.fit_params}
        all_args = {
            "fitter": AutoML,
            "fit_args": args,
            "fit_kwargs": used_kwargs,
            "init_args": self.init_args,
            "init_kwargs": self.init_kwargs,
        }
        fitted = memoizer(fitter_fun, all_args)
        self.__dict__ = fitted.__dict__



from econml.dml import LinearDML
from sklearn.base import clone
import numpy as np
from scipy.special import softmax
from econml.score.ensemble_cate import EnsembleCateEstimator

# code adapted from https://github.com/microsoft/EconML/blob/main/econml/score/rscorer.py


class RScoreWrapper:
    def __init__(
        self,
        model_y,
        model_t,
        train_df,
        test_df,
        outcome,
        treatment,
        common_causes,
        effect_modifiers,
    ):
        self.train = RScorer(model_y=model_y, model_t=model_t, discrete_treatment=True)
        self.test = RScorer(model_y=model_y, model_t=model_t, discrete_treatment=True)
        self.train.fit(
            train_df[outcome].to_numpy(),
            train_df[treatment].to_numpy(),
            train_df[common_causes].to_numpy(),
            train_df[effect_modifiers].to_numpy(),
        )
        self.test.fit(
            test_df[outcome].to_numpy(),
            test_df[treatment].to_numpy(),
            test_df[common_causes].to_numpy(),
            test_df[effect_modifiers].to_numpy(),
        )


class RScorer:
    """Scorer based on the RLearner loss. Fits residual models at fit time and calculates
    residuals of the evaluation data in a cross-fitting manner::

        Yres = Y - E[Y|X, W]
        Tres = T - E[T|X, W]

    Then for any given cate model calculates the loss::

        loss(cate) = E_n[(Yres - <cate(X), Tres>)^2]

    Also calculates a baseline loss based on a constant treatment effect model, i.e.::

        base_loss = min_{theta} E_n[(Yres - <theta, Tres>)^2]

    Returns an analogue of the R-square score for regression::

        score = 1 - loss(cate) / base_loss

    This corresponds to the extra variance of the outcome explained by introducing heterogeneity
    in the effect as captured by the cate model, as opposed to always predicting a constant effect.
    A negative score, means that the cate model performs even worse than a constant effect model
    and hints at overfitting during training of the cate model.

    This method was also advocated in recent work of [Schuleretal2018]_ when compared among several alternatives
    for causal model selection and introduced in the work of [NieWager2017]_.

    Parameters
    ----------
    model_y: estimator
        The estimator for fitting the response to the features. Must implement
        `fit` and `predict` methods.

    model_t: estimator
        The estimator for fitting the treatment to the features. Must implement
        `fit` and `predict` methods.

    discrete_treatment: bool, optional (default is ``False``)
        Whether the treatment values should be treated as categorical, rather than continuous, quantities

    categories: 'auto' or list, default 'auto'
        The categories to use when encoding discrete treatments (or 'auto' to use the unique sorted values).
        The first category will be treated as the control treatment.

    cv: int, cross-validation generator or an iterable, optional (Default=2)
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 3-fold cross-validation,
        - integer, to specify the number of folds.
        - :term:`CV splitter`
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if the treatment is discrete
        :class:`~sklearn.model_selection.StratifiedKFold` is used, else,
        :class:`~sklearn.model_selection.KFold` is used
        (with a random shuffle in either case).

        Unless an iterable is used, we call `split(concat[W, X], T)` to generate the splits. If all
        W, X are None, then we call `split(ones((T.shape[0], 1)), T)`.

    mc_iters: int, optional (default=None)
        The number of times to rerun the first stage models to reduce the variance of the nuisances.

    mc_agg: {'mean', 'median'}, optional (default='mean')
        How to aggregate the nuisance value for each sample across the `mc_iters` monte carlo iterations of
        cross-fitting.

    random_state: int, :class:`~numpy.random.mtrand.RandomState` instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If :class:`~numpy.random.mtrand.RandomState` instance, random_state is the random number generator;
        If None, the random number generator is the :class:`~numpy.random.mtrand.RandomState` instance used
        by :mod:`np.random<numpy.random>`.

    References
    ----------
    .. [NieWager2017] X. Nie and S. Wager.
        Quasi-Oracle Estimation of Heterogeneous Treatment Effects.
        arXiv preprint arXiv:1712.04912, 2017.
        `<https://arxiv.org/pdf/1712.04912.pdf>`_

    .. [Schuleretal2018] Alejandro Schuler, Michael Baiocchi, Robert Tibshirani, Nigam Shah.
        "A comparison of methods for model selection when estimating individual treatment effects."
        Arxiv, 2018
        `<https://arxiv.org/pdf/1804.05146.pdf>`_

    """

    def __init__(
        self,
        *,
        model_y,
        model_t,
        discrete_treatment=False,
        categories="auto",
        cv=2,
        mc_iters=None,
        mc_agg="mean",
        random_state=None
    ):
        self.model_y = clone(model_y, safe=False)
        self.model_t = clone(model_t, safe=False)
        self.discrete_treatment = discrete_treatment
        self.cv = cv
        self.categories = categories
        self.random_state = random_state
        self.mc_iters = mc_iters
        self.mc_agg = mc_agg

    def fit(self, y, T, X=None, W=None, sample_weight=None, groups=None):
        """

        Parameters
        ----------
        Y: (n x d_y) matrix or vector of length n
            Outcomes for each sample
        T: (n x dₜ) matrix or vector of length n
            Treatments for each sample
        X: optional (n x dₓ) matrix
            Features for each sample
        W: optional (n x d_w) matrix
            Controls for each sample
        sample_weight: optional (n,) vector
            Weights for each row
        groups: (n,) vector, optional
            All rows corresponding to the same group will be kept together during splitting.
            If groups is not None, the `cv` argument passed to this class's initializer
            must support a 'groups' argument to its split method.

        Returns
        -------
        self
        """
        if X is None:
            raise ValueError("X cannot be None for the RScorer!")

        self.lineardml_ = LinearDML(
            model_y=self.model_y,
            model_t=self.model_t,
            cv=self.cv,
            discrete_treatment=self.discrete_treatment,
            categories=self.categories,
            random_state=self.random_state,
            mc_iters=self.mc_iters,
            mc_agg=self.mc_agg,
        )
        self.lineardml_.fit(
            y,
            T,
            X=None,
            W=np.hstack([v for v in [X, W] if v is not None]),
            sample_weight=sample_weight,
            groups=groups,
            cache_values=True,
        )
        self.base_score_ = self.lineardml_.score_
        self.dx_ = X.shape[1]
        return self

    def score(self, cate_estimate):
        """
        Parameters
        ----------
        cate_model : instance of fitted BaseCateEstimator

        Returns
        -------
        score : double
            An analogue of the R-square loss for the causal setting.
        """
        Y_res, T_res = self.lineardml_._cached_values.nuisances
        # X = self.lineardml_._cached_values.W[:, : self.dx_]
        sample_weight = self.lineardml_._cached_values.sample_weight
        if Y_res.ndim == 1:
            Y_res = Y_res.reshape((-1, 1))
        if T_res.ndim == 1:
            T_res = T_res.reshape((-1, 1))
        effects = cate_estimate.reshape((-1, Y_res.shape[1], T_res.shape[1]))
        Y_res_pred = np.einsum("ijk,ik->ij", effects, T_res).reshape(Y_res.shape)
        if sample_weight is not None:
            return (
                1
                - np.mean(
                    np.average((Y_res - Y_res_pred) ** 2, weights=sample_weight, axis=0)
                )
                / self.base_score_
            )
        else:
            return 1 - np.mean((Y_res - Y_res_pred) ** 2) / self.base_score_

    def best_model(self, cate_models, return_scores=False):
        """Chooses the best among a list of models

        Parameters
        ----------
        cate_models : list of instances of fitted BaseCateEstimator
        return_scores : bool, optional (default=False)
            Whether to return the list scores of each model
        Returns
        -------
        best_model : instance of fitted BaseCateEstimator
            The model that achieves the best score
        best_score : double
            The score of the best model
        scores : list of double
            The list of scores for each of the input models. Returned only if `return_scores=True`.
        """
        rscores = [self.score(mdl) for mdl in cate_models]
        best = np.nanargmax(rscores)
        if return_scores:
            return cate_models[best], rscores[best], rscores
        else:
            return cate_models[best], rscores[best]

    def ensemble(self, cate_models, eta=1000.0, return_scores=False):
        """Ensembles a list of models based on their performance

        Parameters
        ----------
        cate_models : list of instances of fitted BaseCateEstimator
        eta : double, optional (default=1000)
            The soft-max parameter for the ensemble
        return_scores : bool, optional (default=False)
            Whether to return the list scores of each model
        Returns
        -------
        ensemble_model : instance of fitted EnsembleCateEstimator
            A fitted ensemble cate model that calculates effects based on a weighted
            version of the input cate models, weighted by a softmax of their score
            performance
        ensemble_score : double
            The score of the ensemble model
        scores : list of double
            The list of scores for each of the input models. Returned only if `return_scores=True`.
        """
        rscores = np.array([self.score(mdl) for mdl in cate_models])
        goodinds = np.isfinite(rscores)
        weights = softmax(eta * rscores[goodinds])
        goodmodels = [mdl for mdl, good in zip(cate_models, goodinds) if good]
        ensemble = EnsembleCateEstimator(cate_models=goodmodels, weights=weights)
        ensemble_score = self.score(ensemble)
        if return_scores:
            return ensemble, ensemble_score, rscores
        else:
            return ensemble, ensemble_score



from typing import Any, Union, Sequence, Optional
import math

import numpy as np
import pandas as pd

from causaltune.memoizer import MemoizingWrapper


def clean_config(params: dict):
    # TODO: move this to formal constraints in tune?
    if "subforest_size" in params and "n_estimators" in params:
        params["n_estimators"] = params["subforest_size"] * math.ceil(
            params["n_estimators"] / params["subforest_size"]
        )

    if "min_samples_split" in params and params["min_samples_split"] > 1.5:
        params["min_samples_split"] = int(params["min_samples_split"])
    return params


def treatment_values(treatment: pd.Series, control_value: Any):
    return sorted([t for t in treatment.unique() if t != control_value])


def treatment_is_multivalue(treatment: Union[int, str, Sequence]) -> bool:
    if isinstance(treatment, str) or len(treatment) == 1:
        return False
    else:
        return True


#
# def fit_params_wrapper(parent: type):
#     class FitParamsWrapper(parent):
#         def __init__(self, *args, fit_params=None, **kwargs):
#             self.init_args = args
#             self.init_kwargs = kwargs
#             if fit_params is not None:
#                 self.fit_params = fit_params
#             else:
#                 self.fit_params = {}

#         def fit(self, *args, **kwargs):
#             # we defer the initialization to the fit() method so we can memoize the fit
#             # using all the args from both init and fit
#             used_kwargs = {**kwargs, **self.fit_params}
#             to_hash = {
#                 "class": super().__class__.__name__,
#                 "fit_args": args,
#                 "fit_kwargs": used_kwargs,
#                 "init_args": self.init_args,
#                 "init_kwargs": self.init_kwargs,
#             }
#             test_fun(to_hash)

#             super().__init__(*self.init_args, **self.init_kwargs)

#             print("calling AutoML fit method with ", used_kwargs)
#             super().fit(*args, **used_kwargs)

#     return FitParamsWrapper


AutoMLWrapper = MemoizingWrapper


# AutoMLWrapper = fit_params_wrapper(AutoML)

# class AutoMLWrapper(AutoML):
#     def __init__(self, *args, fit_params=None, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.init_args = args
#         self.init_kwargs = kwargs
#         if fit_params is not None:
#             self.fit_params = fit_params
#         else:
#             self.fit_params = {}
#
#     def fit(self, *args, **kwargs):
#         # we defer the initialization to the fit() method so we can memoize the fit
#         # using all the args from both init and fit
#         used_kwargs = {**kwargs, **self.fit_params}
#         print("calling AutoML fit method with ", used_kwargs)
#         super().fit(*args, **used_kwargs)


def policy_from_estimator(est, df: pd.DataFrame):
    # must be done just like this so it also works for metalearners
    X_test = df[est.estimator._effect_modifier_names]
    return est.estimator.estimator.effect(X_test) > 0


def generate_psdmat(n_dims: int = 10) -> np.ndarray:
    """generates a symmetric, positive semidefinite matrix

    Args:
        n_dims (int, optional): number of dimensions. Defaults to 10.

    Returns:
        np.ndarray: psd matrix
    """
    A = np.random.rand(n_dims, n_dims)
    A = A @ A.T

    return A


def psw_joint_weights(a: np.ndarray, b: Optional[np.ndarray] = None):
    """Generate inverse joint propensity-score weights from two vectors

    Args:
        a (np.ndarray): propensity score vector of length n_A, array of shape (n_A,) or (n_A, 1)
        b (Optional[np.ndarray], optional): propensity score vector of length n_B,
            array of shape (n_B,) or (n_B, 1).
            If None, compares a with itself. Defaults to None.Defaults to None.

    Returns:
        (np.ndarray): inverse propensity score product matrix of shape (n_A, n_B)
    """
    if b is None:
        b = a

    assert a.ndim < 3 and b.ndim < 3

    if a.ndim == 1:
        a = np.expand_dims(a, axis=1)
    if b.ndim == 1:
        b = np.expand_dims(b, axis=1)

    assert a.shape[1] == b.shape[1]

    w = np.multiply(a, b.T)
    ipw = np.reciprocal(w)

    return ipw
